/************************************************************************
*									*
*	gimp_dilate.c							*
*	This file contains the code necessary to perform a grayscale	*
*	dilation on a image for a GIMP plug-in.				*
*									*
*	Author: Bryan Evenson						*
*									*
*		-Morphological dilation adapted from a 			*
*		MATLAB version written by Yali Wei (mdilate.m), 	*
*		available at http:/www.mathworks.com/			*
*									*
*		-Ideas for the GIMP plug-in specific parts of the	*
*		code liberally borrowed from files at 	 		*
*		http:/cvs.gnome.org.					*
*									*
*									*
*	Created: May 6, 1999						*
*									*
*	Revision 1.6							*
*									*	
************************************************************************/

#include <stdlib.h>
#include <libgimp/gimp.h>
#include "gimp_morph.h"

GPixelRgn gimp_dilate(GDrawable *drawable, int Se[5][5], int Px, int Py)
{
	GPixelRgn srcPR, destPR;
	gint width, height;
	gint bytes;
	guchar *dest;
	guchar *src_row[5], *tmp;
	gint row, col, tmpMax, tmpVal, dil_col, dil_row;
	gint x1, y1, x2, y2;
	gint m, n;
    

	/* Get the input area. This is the bounding box of the selection in
	*  the image (or the entire image if there is no selection). Only
	*  operating on the input area is simply an optimization. It doesn't
	*  need to be done for correct operation. (It simply makes it go
	*  faster, since fewer pixels need to be operated on).
	*/
	gimp_drawable_mask_bounds (drawable->id, &x1, &y1, &x2, &y2);
	gimp_progress_init ("Computing dilation...");

	/* Get the size of the input image. (This will/must be the same
	 *  as the size of the output image.
	 */
	width = drawable->width;
	height = drawable->height;
	bytes = drawable->bpp;

	/*  allocate row buffers  */
	for(row=0; row<5; row++)
		src_row[row] = (guchar *) malloc ((x2 - x1) * bytes);
	dest = (guchar *) malloc ((x2 - x1) * bytes);


	/*  initialize the pixel regions  */
	gimp_pixel_rgn_init (&srcPR, drawable, 0, 0, width, height, FALSE, FALSE);
	gimp_pixel_rgn_init (&destPR, drawable, 0, 0, width, height, TRUE, TRUE);

	/* Properly initalize the row buffers. */
	for(row=y1; row<y1+Py; row++)
		memset(src_row[row], 0, sizeof(src_row[row]));

	for(row=y1+Py; row<y1+5; row++)
		gimp_pixel_rgn_get_row (&srcPR, src_row[row], x1, row-Py, (x2 - x1));
	
	
	/*  loop through the rows  */
	for (row = y1; row < y2; row++)
	{

	  /*  prepare the end row of the buffer */
	  if(row<y2-(4-Py))
		gimp_pixel_rgn_get_row(&srcPR, src_row[4], x1, row+(4-Py), (x2 - x1));
	  else
		  memset(src_row[4], 0, sizeof(src_row[4]));

	  /* Fill in data for the row. */
	  for (col = x1; col < x2; col++)
	  {
	    /*Run the mask over the current pixel.  Whichever value is the greatest */
	    /*Becomes the new pixel value. */
	    tmpMax = 0;
	    m = 0;
	    while(m<5)
	    {
	      n = 0;
	      while(n<5 && tmpMax != 255)
	      {
		dil_col=col-Px+n;
               	dil_row=m;

               	/*Check to make sure the point is within the bounds */
               	/*of the buffer. */
               	if(dil_col>=0 && dil_col<x2)
               	{
               	        /*If the new value is smaller than the current smallest */
               	        /*value, replace tmpMax */
               		tmpVal = (gint)src_row[dil_row][dil_col*bytes]+(gint)Se[m][n];

                       	/*Since dest[] is of the type (guchar *), make sure that it */
                       	/*doesn't get a signed negative value. */
                       	if(tmpVal>tmpMax)
                       	        tmpMax = (tmpVal>255 || tmpVal<0) ? 255 : tmpVal;
                }
		n++;
	      }
	      m++;
	    }
	    
	    /*Store the new pixel value. */
	    dest[col*bytes] = tmpMax;
	  }

	  /*  store the dest  */
	  gimp_pixel_rgn_set_row (&destPR, dest, x1, row, (x2 - x1));

	  /*  shuffle the row pointers  */
	  tmp = src_row[0];
	  src_row[0] = src_row[1];
	  src_row[1] = src_row[2];
	  src_row[2] = src_row[3];
	  src_row[3] = src_row[4];
	  src_row[4] = tmp;

	  /*Update the progress meter. */
	  if ((row % 5) == 0)
	    gimp_progress_update ((double) row / (double) (y2 - y1));
	} 

	for(row=0; row<5; row++)
		free (src_row[row]);
	free (dest);

	return destPR;
}
